#!/usr/bin/tclsh

#########################################################################
# This represents a low-level library for communication with
# a Slink-e device as manufactured by nirvis (http://www.nirvic.com).
# We implemented this library for our own needs and the purpose
# of integration in some of our demos.
# The functionality of the Slink-E is almost completely represented
# in the API. Several functions that can be found here and start
# with an underline character ('_') are meant for internal use only
# and should not be called by an application programmer.
# Before one starts using this library one has to call the following
# two functions:
#  - slinke::init  (initializes this library)
#  - slinke::connect <serial device name, i.e. "/dev/ttyS0">
# At the end the following function should be called to
# release the connection to the Slink-E over the serial port.
#  - slinke::disconnect
#
# Some notes:
#  While you are connected to the Slink-E through this library
#  all IR and parallel port receive messages are disabled to 
#  prevent buffer overflow and other bad things from happening. 
#  Only when commands regarding these ports are used, the port 
#  communication is activated.
#  If you leave the library with a slinke::disconnect the
#  ports will automatically be reactivated and a baudrate of 38400
#  restored on the Slink-E device.
#########################################################################

source local.tcl

#-----
# TODO:
# - IR routing command
# - timeout for receiving (?)
# - Receiving of data also from parallel port (?)
#-----

namespace eval Slinke {
  ## for commands  and their parameters
  variable a_commands      
  ## for responses and their parameters
  variable a_responses     
  ## the names of the responses 
  variable l_responsenames 
  ## serial port file descriptor
  variable fd            
  ## Name of the serial port
  variable serialport
  ## baudrate used on serial port
  variable baudrate
  ## the buttons and their msg sequences
  variable a_buttons       
  ## the names of the buttons
  variable l_buttonnames   
}


#--------------------------------------------------------
# Slinke::init  --- initialize the slinke library
#
# Function called to initialize this library.
# You must call this function before doing anything!
#--------------------------------------------------------
proc Slinke::init {} {
  set IR     4
  set Par    5
  set Serial 6
  set System 7

  set Slinke::l_responsenames ""
  #Create a couple of commands that we will need here
  #Attention: commands and responses with "reception" in their
  #           name are treated special! Only those commands 
  #           enabling or disabling port communication on a
  #           certain port should have this word in their name!

  #####################
  #Serial Port related.
  Slinke::new_command  get_baudrate          $Serial   9   0  none
  Slinke::new_response get_baudrate          $Serial   8   1  none
  Slinke::new_response invalid_baudrate      $Serial  32   0  none

  Slinke::new_command  set_baudrate          $Serial   8   1  none

  #######################
  #Parallel Port related.  
  Slinke::new_command  get_handshaking       $Par     17   0  Par
  Slinke::new_response get_handshaking       $Par     16   1  Par

  Slinke::new_command  set_handshaking       $Par     16   1  Par


  Slinke::new_command  get_direction         $Par     19   0  Par
  Slinke::new_response get_direction         $Par     18   1  Par

  Slinke::new_command  set_direction         $Par     18   1  Par

  Slinke::new_command  sample_port           $Par     20   0  Par

  Slinke::new_command  par_enable_reception  $Par      3   0  Par
  Slinke::new_response par_enable_reception  $Par      3   0  Par

  Slinke::new_command  par_disable_reception $Par      2   0  Par
  Slinke::new_response par_disable_reception $Par      2   0  Par

  ############
  #IR related.
  Slinke::new_response last_received_port    $IR       1   1  IR

  Slinke::new_command  get_sample_period     $IR       5   0  IR
  Slinke::new_response get_sample_period     $IR       4   2  IR
  Slinke::new_response invalid_sample_period $IR     130   0  IR

  Slinke::new_command  set_sample_period     $IR       4   2  IR


  Slinke::new_command  get_ir_carrier        $IR       7   0  IR
  Slinke::new_response get_ir_carrier        $IR       6   2  IR

  Slinke::new_command  set_ir_carrier        $IR       6   2  IR


  Slinke::new_command  get_ir_timeout        $IR      13   0  IR
  Slinke::new_response get_ir_timeout        $IR      12   2  IR

  Slinke::new_command  set_ir_timeout        $IR      12   2  IR

  Slinke::new_command  get_min_ir_msg_len    $IR      15   0  IR
  Slinke::new_response get_min_ir_msg_len    $IR      14   1  IR

  Slinke::new_command  set_min_ir_msg_len    $IR      14   1  IR


  Slinke::new_command  get_ir_tx_ports       $IR      19   0  IR
  Slinke::new_response get_ir_tx_ports       $IR       8   1  IR

  Slinke::new_command  set_ir_tx_ports       $IR       8   1  IR


  Slinke::new_command  get_ir_rx_ports       $IR      18   0  IR
  Slinke::new_response get_ir_rx_ports       $IR       9   1  IR

  Slinke::new_command  set_ir_rx_ports       $IR       9   1  IR


  Slinke::new_command  get_ir_rx_port_pol    $IR      17   0  IR
  Slinke::new_response get_ir_rx_port_pol    $IR      11   1  IR

  Slinke::new_command  set_ir_rx_port_pol    $IR      11   1  IR

  Slinke::new_command  get_ir_routing        $IR      16   0  IR
  Slinke::new_response get_ir_routing        $IR      10  16  IR

  Slinke::new_command  set_ir_routing        $IR      10  16  IR

  Slinke::new_command  ir_enable_reception   $IR       3   0  IR
  Slinke::new_response ir_enable_reception   $IR       3   0  IR

  Slinke::new_command  ir_disable_reception  $IR       2   0  IR
  Slinke::new_response ir_disable_reception  $IR       2   0  IR

  ################
  #System related.
  Slinke::new_command  get_version           $System  11  0   none
  Slinke::new_response get_version           $System  11  1   none

  Slinke::new_command  get_serial            $System  12  0   none
  Slinke::new_response get_serial            $System  12  8   none

  Slinke::new_command  load_settings         $System  15  0   none
  Slinke::new_response load_settings         $System  15  0   none

  Slinke::new_command  save_settings         $System  14  0   none
  Slinke::new_response save_settings         $System  14  0   none
  Slinke::new_response eeprom_write_error    $System 143  0   none

  Slinke::new_command  cmd_resume            $System 170  0   none

  Slinke::new_command  all_enable_reception  $System   3   0  System
  Slinke::new_response all_enable_reception  $System   3   0  System

  Slinke::new_command  all_disable_reception $System   2   0  System
  Slinke::new_response all_disable_reception $System   2   0  System

  ################
  # Error messages
  Slinke::new_response serial_recv_overflow  $Serial 131   0  none
  Slinke::new_response serial_framing_error  $Serial 133   0  none  
  Slinke::new_response serial_recv_overrrun  $Serial 134   0  none
#  Slinke::new_response transmit_timeout      any     129   0  none
}

#--------------------------------------------------------
# Function called to low-level connect to the Slink-e 
# device.
#--------------------------------------------------------
proc Slinke::connect {devname {baudrate 38400}} {
  if {"Linux" == $Local::os} then {
    exec /bin/stty -F $devname $baudrate crtscts
  }
  set Slinke::fd [open $devname r+]
  set Slinke::baudrate   $baudrate
  set Slinke::serialport $devname
  puts "Connected to Slink-E on $devname at $baudrate baud."
  fconfigure $Slinke::fd -mode $baudrate,n,8,1 \
                         -blocking 1 \
                         -buffersize 40 \
                         -translation binary
  # Override any previous error condition.
  Slinke::do_cmd  cmd_resume none none
  Slinke::disable_reception "IR"
}


#------------------------------------------------------
# Disconnect from the serial port -s hould be called at
# the very end of using this library.
#------------------------------------------------------
proc Slinke::disconnect {{enable true}} {
  if {"true" == $enable} then {
    Slinke::enable_reception "System"
    Slinke::set_baudrate 38400
  }
  close $Slinke::fd
  set Slinke::fd 0
}


#--------------------------------------------------
# Enable a given port for data reception
#--------------------------------------------------
proc Slinke::enable_reception {portname} {
  set type ""
  switch $portname {
    "IR"     {set type  ir_enable_reception}
    "Par"    {set type par_enable_reception}
    "System" {set type all_enable_reception}
  }

  if {"" != $type} {
    Slinke::do_cmd $type $type none
    return "OK"
  }
  return "Unknown portname $portname!"
}


#--------------------------------------------------
# Disable a given port for data reception
#--------------------------------------------------
proc Slinke::disable_reception {portname} {
  set type ""
  switch $portname {
    "IR"     {set type  ir_disable_reception}
    "Par"    {set type par_disable_reception}
    "System" {set type all_disable_reception}
  }

  if {"" != $type} {
    Slinke::do_cmd $type $type none
    return "OK"
  }
  return "Unknown portname $portname!"
}


#--------------------------------------------------
# Add a new command to the list of commands the
# system knows.
#--------------------------------------------------
proc Slinke::new_command {args} {
  set tag [lindex $args 0]
  set Slinke::a_commands($tag) [lrange $args 1 end]
}


#--------------------------------------------------
# Add a new response to the list of repsonses the
# system needs to know.
#--------------------------------------------------
proc Slinke::new_response {args} {
  set tag [lindex $args 0]
  set Slinke::a_responses($tag) [lrange $args 1 end]
  if {-1 == [lsearch -exact $Slinke::l_responsenames $tag]} then {
    lappend Slinke::l_responsenames $tag
  }
}


#-----------------------------------------------------------
# Execute a command and if specified also try to get the
# response.
#-----------------------------------------------------------
proc Slinke::do_cmd {tag response error args} {
  if {"none" != $tag} then {
    set cmd $Slinke::a_commands($tag)
    #Enable reception on this port, i.e. "IR"
    if {-1 == [string first "reception" $tag]} then {
      Slinke::enable_reception [lindex $cmd 3]
    }
    set c0       "[expr ([lindex $cmd 0] << 5) | 31]"
    set c1       "[expr  [lindex $cmd 1]           ]"
    set num_args "[expr  [lindex $cmd 2]           ]"
    set cmd [format "%c%c" $c0 $c1]
    if {"" != $Slinke::fd} then {
      for {set i 0} {$i < $num_args} {incr i} {
        append cmd [format "%c" [lindex $args $i]]
      }
      puts -nonewline $Slinke::fd $cmd
      flush $Slinke::fd
    } else {
      puts "You must call Slinke::connect first!"
    }
  }
  if {"none" != $response} then {
    set resp [do_resp $response $error]
    #Disable reception on this port, i.e. "IR"
    if {-1 == [string first "reception" $tag]} then {
      Slinke::disable_reception [lindex $cmd 3]
    }
    return $resp
  }
}


#-----------------------------------------------------------
# Get a response
#-----------------------------------------------------------
proc Slinke::do_resp {response err_response} {

  #If I got a response back then let me
  #check whether this is the response that I 
  #expect
  while {1} {
    #First read the header bytes...
    set resp [Slinke::_read_header]
    #resp: port len c1 "resp"
    if {"resp" == [lindex $resp 3]} then {

      set exp_resp $Slinke::a_responses($response)
      # The third value is the number of additional bytes
      # that I expect.
      set _port     [expr [lindex $exp_resp 0]]
      set _c1       [expr [lindex $exp_resp 1]]
      if {$_c1   == [lindex $resp 2] &&
          $_port == [lindex $resp 0]} then {
        # In this case everything went alright
        set num_args  [lindex $exp_resp 2]
        set inp [read $Slinke::fd $num_args]
        return $inp
      } elseif {"none" != $err_response} {
        #Let me see whether this is an expected error
        set exp_resp $Slinke::a_responses($err_response)
        # The third value is the number of additional bytes
        # that I expect.
        set _port     [expr [lindex $exp_resp 0]]
        set _c1       [expr [lindex $exp_resp 1]]
        if {$_c1   == [lindex $resp 2] &&
            $_port == [lindex $resp 0]} then {
          set num_args [lindex $exp_resp 2]
          if {0 != $num_args} then {
            set inp [read $Slinke::fd $num_args]
          }
          #return "invalid parameter"
        }
      } else {
        #something went really wrong in this case
        if {1 != [lindex $resp 2]} then {
          puts "c1=[lindex $resp 2], port= [lindex $resp 0], $response, $err_response"
          Slinke::do_cmd cmd_resume none none
          # Sometimes one more end of reception comes along...
          break
        }
        puts "c1=[lindex $resp 2] len=[lindex $resp 1] [lindex $resp 3]"
      }
    } elseif {"data end" == [lindex $resp 3]} then {
    } else {
      ## This can only be a "recv data". I just discard
      ## the data here.
      read $Slinke::fd [expr [lindex $resp 1]]
    }
  }
  return ""
}


#------------------------------------------------------------
# Read the version information from the slink-e and return
# a formatted string.
#------------------------------------------------------------
proc Slinke::get_version {} {
  set ver [Slinke::do_cmd get_version get_version none]
  if {1 == [string length $ver]} then {
    scan $ver "%c" version
    return [format "%u.%u" [expr $version >> 4] [expr $version & 15]]
  }
  return "error"
}


#------------------------------------------------------------
# Read the serial number of the slink-e and return a formatted
# string.
#------------------------------------------------------------
proc Slinke::get_serial {} {
  set ser [Slinke::do_cmd get_serial get_serial none]
  if {8 == [string length $ser]} then {
    scan $ser "%c%c%c%c%c%c%c%c" s0 s1 s2 s3 s4 s5 s6 s7
    return [format "%02x%02x%02x%02x%02x%02x%02x%02x" $s0 $s1 $s2 $s3 $s4 $s5 $s6 $s7]
  }
  return "error"
}


#------------------------------------------------------------
# Load settings from EEprom. Beware, this might also change
# the baudrate.
#------------------------------------------------------------
proc Slinke::load_settings {} {
  Slinke::do_cmd load_settings load_settings none
  return "Ok"
}


#------------------------------------------------------------
# Save settings to EEprom. Beware, this will also store the
# baudrate.
#------------------------------------------------------------
proc Slinke::save_settings {} {
  Slinke::do_cmd save_settings save_settings eeprom_write_error
  return "Ok"
}


#-----------------------------------------------------------
# Get the currently used baudrate from the slink-e
#-----------------------------------------------------------
proc Slinke::get_baudrate {} {
  set baud [Slinke::do_cmd get_baudrate get_baudrate none]
  if {1 == [string length $baud]} then {
    scan $baud "%c" b0
    set rc "unknown"
    switch $b0 {
      0 {set rc "2400" }
      1 {set rc "4800" }
      2 {set rc "9600" }
      3 {set rc "19200"}
      4 {set rc "38400"}
    }
    return $rc
  } 
  return "error"
}


#-----------------------------------------------------------
# Set the baudrate of the slink-e
#-----------------------------------------------------------
proc Slinke::set_baudrate {baudrate} {
  set br 4
  switch $baudrate {
    2400  {set br 0 }
    4800  {set br 1 }
    9600  {set br 2 }
    19200 {set br 3 }
    38400 {set br 4 }
  }

  puts "Changing baudrate now."
  set baud [Slinke::do_cmd set_baudrate get_baudrate none $br]
  if {1 == [string length $baud]} then {
    if {$Slinke::baudrate != $br} then {
      Slinke::disconnect false
      Slinke::connect $Slinke::serialport $baudrate
      #puts "Trying to read on more byte after changing baudrate."
      #read $Slinke::fd 1
    }
    scan $baud "%c" br
    return [format "0x%x" [expr $br]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the handshaking mode of the parallel port
#-----------------------------------------------------------
proc Slinke::get_handshaking {} {
  set hm [Slinke::do_cmd get_handshaking get_handshaking none]
  if {1 == [string length $hm]} then {
    scan $hm "%c" phm
    return [format "0x%x" [expr $phm]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the handshaking mode:
#
#-----------------------------------------------------------
proc Slinke::set_handshaking {hm} {
  set phm [Slinke::do_cmd set_handshaking get_handshaking none $hm]
  if {1 == [string length $phm]} then {
    scan $phm "%c" _hm
    return [format "0x%x" [expr $_hm]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the direction of the parallel port
#-----------------------------------------------------------
proc Slinke::get_direction {} {
  set d [Slinke::do_cmd get_direction get_direction none]
  if {1 == [string length $d]} then {
    scan $d "%c" _d
    return [format "0x%x" [expr $_d]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the direction of the parallel port
#-----------------------------------------------------------
proc Slinke::set_direction {d} {
  set pd [Slinke::do_cmd set_direction get_direction none $d]
  if {1 == [string length $pd]} then {
    scan $pd "%c" _d
    return [format "0x%x" [expr $_d]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the currently used sample period from the Slink-E
#-----------------------------------------------------------
proc Slinke::get_ir_sample_period {} {
  set sp [Slinke::do_cmd get_sample_period get_sample_period none]
  if {2 == [string length $sp]} then {
    scan $sp "%c%c" sh sl
    return [format "%u" [expr $sh * 256 + $sl]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the sample period on the Slink-E
#-----------------------------------------------------------
proc Slinke::set_ir_sample_period {sp} {
  set sh [expr $sp / 256]
  set sl [expr $sp & 255]
  set sp [Slinke::do_cmd set_sample_period get_sample_period invalid_sample_period $sh $sl]
  if {2 == [string length $sp]} then {
    scan $sp "%c%c" sh sl
    return [format "%u" [expr $sh * 256 + $sl]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the IR transmission port. Each one of the 8 bits
# turns a transmission port on when set.
#-----------------------------------------------------------
proc Slinke::set_ir_transmit_ports {portbits} {
  Slinke::do_cmd set_ir_tx_ports none none $portbits
}


#-----------------------------------------------------------
# Get the currently currently used IR carrier frequency
# from the slink-e.
#-----------------------------------------------------------
proc Slinke::get_ir_carrier {} {
  set ic [Slinke::do_cmd get_ir_carrier get_ir_carrier none]
  if {2 == [string length $ic]} then {
    scan $ic "%c%c" cp cc
    return [format "%g" [expr pow(2,$cp)*($cc+1)*0.000002 ]]
  }
}


#-----------------------------------------------------------
# Set the IR carrier frequency to be used on the slink-e.
# You must pass a prescaler value as the first vaiable and
# the carrier value as the second parameter.
#-----------------------------------------------------------
proc Slinke::set_ir_carrier {cp cc} {
  set _cp [format "%c" $cp]
  set _cc [format "%c" $cc]
  set ic [Slinke::do_cmd set_ir_carrier get_ir_carrier none $_cp $_cc]
  if {2 == [string length $ic]} then {
    scan $ic "%c%c" cp cc
    return [format "%g" [expr (2^$cp)*($cc+1)*0.2 ]]
  }
}


#-----------------------------------------------------------
# Get the IR timeout
#-----------------------------------------------------------
proc Slinke::get_ir_timeout {} {

  set ito [Slinke::do_cmd get_ir_timeout get_ir_timeout none]
  if {2 == [string length $ito]} then {
    scan $ito "%c%c" th tl
    return [format "%u" [expr $th * 256 + $tl ]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the IR timeout
#-----------------------------------------------------------
proc Slinke::set_ir_timeout {to} {
  set th [expr $to / 256]
  set tl [expr $to & 255]
  set timeout [Slinke::do_cmd set_ir_timeout get_ir_timeout none $th $tl]
  if {2 == [string length $timeout]} then {
    scan $timeout "%c%c" th tl
    return [format "%u" [expr $th * 256 + $tl]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the IR timeout
#-----------------------------------------------------------
proc Slinke::get_ir_timeout {} {

  set ito [Slinke::do_cmd get_ir_timeout get_ir_timeout none]
  if {2 == [string length $ito]} then {
    scan $ito "%c%c" th tl
    return [format "%u" [expr $th * 256 + $tl ]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the minimum IR message length the slink-e will send
# towards the host.
#-----------------------------------------------------------
proc Slinke::get_min_ir_msg_len {} {
  set iml [Slinke::do_cmd get_min_ir_msg_len get_min_ir_msg_len none]
  if {1 == [string length $iml]} then {
    scan $iml "%c" mml
    return [format "%u" [expr $mml]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the minimum IR message length the slink-e will send
# towards the host.
#-----------------------------------------------------------
proc Slinke::set_min_ir_msg_len {mml} {
  set iml [Slinke::do_cmd set_min_ir_msg_len get_min_ir_msg_len none $mml]
  if {1 == [string length $iml]} then {
    scan $iml "%c" mml
    return [format "%u" [expr $mml]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the currently used IR transmission ports
#-----------------------------------------------------------
proc Slinke::get_ir_transmit_ports {} {
  set itp [Slinke::do_cmd get_ir_tx_ports get_ir_tx_ports none]
  if {1 == [string length $itp]} then {
    scan $itp "%c" tp
    return [format "0x%x" [expr $tp]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the currently used IR receive ports
#-----------------------------------------------------------
proc Slinke::get_ir_receive_ports {} {
  set irp [Slinke::do_cmd get_ir_rx_ports get_ir_rx_ports none]
  if {1 == [string length $irp]} then {
    scan $irp "%c" rp
    return [format "0x%x" [expr $rp]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the IR receive ports
#-----------------------------------------------------------
proc Slinke::set_ir_receive_ports {irrp} {
  set irp [Slinke::do_cmd set_ir_rx_ports get_ir_rx_ports none $irrp]
  if {1 == [string length $irp]} then {
    scan $irp "%c" rp
    return [format "0x%x" [expr $rp]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the currently used IR receive port polarities
#-----------------------------------------------------------
proc Slinke::get_ir_receive_port_polarities {} {
  set irpp [Slinke::do_cmd get_ir_rx_port_pol get_ir_rx_port_pol none]
  if {1 == [string length $irpp]} then {
    scan $irpp "%c" rpp
    return [format "0x%x" [expr $rpp]]
  }
  return "error"
}


#-----------------------------------------------------------
# Set the IR receive port polarities
#-----------------------------------------------------------
proc Slinke::set_ir_receive_port_polarities {irpp} {
  set irpp [Slinke::do_cmd set_ir_rx_port_pol get_ir_rx_port_pol none $irpp]
  if {1 == [string length $irpp]} then {
    scan $irpp "%c" rpp
    return [format "0x%x" [expr $rpp]]
  }
  return "error"
}


#-----------------------------------------------------------
# Get the IR routing table
#-----------------------------------------------------------
proc Slinke::get_ir_routing {} {
  set irrt [Slinke::do_cmd get_ir_routing get_ir_routing none]
  if {16 == [string length $irrt]} then {
    scan $irrt "%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c" irrx0rl irrx0c \
                                                  irrx1rl irrx1c \
                                                  irrx2rl irrx2c \
                                                  irrx3rl irrx3c \
                                                  irrx4rl irrx4c \
                                                  irrx5rl irrx5c \
                                                  irrx6rl irrx6c \
                                                  irrx7rl irrx7c
    return [format "0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x" \
                   $irrx0rl $irrx0c \
                   $irrx1rl $irrx1c \
                   $irrx2rl $irrx2c \
                   $irrx3rl $irrx3c \
                   $irrx4rl $irrx4c \
                   $irrx5rl $irrx5c \
                   $irrx6rl $irrx6c \
                   $irrx7rl $irrx7c]
  }
}


#-----------------------------------------------
# Receive the header of a message. The header
# are only the first few bytes.
#-----------------------------------------------
proc Slinke::_read_header {} {
#puts "Trying to read header."
  set cs0 [read $Slinke::fd 1]
#puts "Got first byte."
  scan $cs0 "%c" c0
  set port [expr $c0 >> 5]
  set len  [expr $c0 & 0x1f]
#  puts "Got len $len"
  if {0 == $len} then {
#puts "data_end"
    return [list $port 0 {} "data end"]
  } elseif {31 == $len} then {
    set cs1 [read $Slinke::fd 1]
    scan $cs1 "%c" c1
#puts "resp"
    # To compensate for a bug(?)
    if {1 == $c1} then {
      return [list $port $len $c1 "data end"]
    }
    return [list $port $len $c1 "resp"]
  } else {
#puts "recv data"
    return [list $port $len {} "recv data"]
  }
}


#-------------------------------------------------
# Receive a message as for example received
# from the IR detector when a IR device is pointed
# to it.
#-------------------------------------------------
proc Slinke::_get_port_msg {} {
  set ret  [Slinke::_read_header]
  set port [lindex $ret 0]
  set len  [lindex $ret 1]
  set type [lindex $ret 3]
  set msg ""
  if {"recv data" == $type} then {
    #puts "Receiving data with len [expr $len]"
    set msg [read $Slinke::fd [expr $len]]
  } elseif {"resp" == $type} then {
    set c1 [lindex $ret 2]
    if {$port == [lindex $Slinke::a_responses(last_received_port) 0] &&
        $c1   == [lindex $Slinke::a_responses(last_received_port) 1]} then {
      set _len   [lindex $Slinke::a_responses(last_received_port) 2]
    } else {
      set _len 0
      puts ""
    }
    read $Slinke::fd $_len
  } elseif {"data end" == $type} then {
    #puts "End of data received."
    Slinke::_get_port_msg
  } else {
    #puts "ERROR!"
  }
  return [list $port $msg]
}


#--------------------------------------------------------
# Receive a message on a given port. This function 
# will only return messages received from the given port.
# (Any other messages should not have been received
# anyway...)
#--------------------------------------------------------
proc Slinke::_receive_data {port} {
  # If it's from parallel port then I need to activate
  # sampling first.
  if {5 == $port} then {
    Slinke::do_cmd sample_port none none
  }
  set msg [Slinke::_get_port_msg]
  if {$port == [lindex $msg 0]} then {
    return [lindex $msg 1]
  }
  return ""
}


#--------------------------------------------------------
# Converts a port name into a port number
#--------------------------------------------------------
proc Slinke::_getport {portname} {
  switch $portname {
    "CTRL0" {set port 0}
    "CTRL1" {set port 1}
    "CTRL2" {set port 2}
    "CTRL3" {set port 3}
    "IR"    {set port 4}
    "PAR"   {set port 5}
    "SER"   {set port 6}
    "SYS"   {set port 7}
  }
}


#--------------------------------------------------------
# Learn a sequence on the given port. 
# The port must be specified by its name. Valid
# names are: "IR"
#--------------------------------------------------------
proc Slinke::learn {portname} {
  puts "learning now!"
  set rc ""
  set port [Slinke::_getport $portname]
  Slinke::enable_reception $portname
  while {0 != $port} {
    set inp [Slinke::_receive_data $port]
    if {"" == $inp} then {
      break
    }
    append rc $inp
  }
  Slinke::disable_reception $portname
  return $rc
}


#--------------------------------------------------------
# Send a given message over the given port, identified
# by its number. The 2nd parameter indicates the length
# of the message to send and must be less or equal
# to the actual length of the message 'msg'.
#--------------------------------------------------------
proc Slinke::sendmsg {port len msg} {
  set c0 [expr $port << 5 | $len]
  set first [format "%c" $c0]
  puts -nonewline $Slinke::fd $first
  flush $Slinke::fd  
  if {0 < [string length $msg]} then {
    puts -nonewline $Slinke::fd $msg
  }
  flush $Slinke::fd  
}


#--------------------------------------------------------
# Send a given message to the given port
# The portname maye be one of the following: "IR"
# The message must be given in form of a string containing
# a sequence of bytes that will be sent over IR for example.
#
# If the message is to be sent over more than one port
# at the same time, the the third paramter can specify
# the outgoing portbits (default is port 0)
#--------------------------------------------------------
proc Slinke::play {portname msg {portbits 0x01}} {
  Slinke::enable_reception $portname
  set port [Slinke::_getport $portname]
  Slinke::do_cmd cmd_resume none none
  Slinke::set_ir_transmit_ports $portbits
  set len [expr [string length $msg]]
  set start 0
  while {$len > 0} {
    set slen $len
    if {$slen > 30} then {
      set slen 30
    }
    Slinke::sendmsg $port $slen [string range $msg $start [expr $start + $slen - 1]]

    set start [expr $start + $slen]
    set len [expr $len - $slen]
  }
  Slinke::sendmsg $port 0 ""
  Slinke::disable_reception $portname
}
 

#--------------------------------------------------------
# Store a sequence of bytes under a given name. The
# idea is that each message can later on be recalled
# using this name (pressing a button on the remote
# control).
#--------------------------------------------------------
proc Slinke::store {button msg {namespace "default-"}} {
  set Slinke::a_buttons($namespace$button) $msg
  set idx [lsearch -exact Slinke::l_buttonnames $namespace$button]
  if {-1 == $idx} then {
    lappend Slinke::l_buttonnames $namespace$button
  }
}


#--------------------------------------------------------
# Play back a message. The message must have been
# stored under a name before. Pass that name (button) to 
# this function and the portname where to play the
# data on (i.e. "IR"). If you want this command to
# be repeated or sent longer then you may also
# pass the number of repetitions you want to this
# function (default is 1).
#--------------------------------------------------------
proc Slinke::recall {button portname {repetitions 1} {namespace "default-"}} {
  for {set i 0} {$i < $repetitions} {incr i} {
    if {1 == [info exists Slinke::a_buttons($namespace$button)]} then {
      Slinke::play $portname $Slinke::a_buttons($namespace$button)
    } else {
      puts "Unknown button."
    }
  }
}


#---------------------------------------------------------------------
# Displays all currently available buttons
#---------------------------------------------------------------------
proc Slinke::showbuttons {} {
  puts "The following buttons are available:"
  puts "$Slinke::l_buttonnames"
}


#---------------------------------------------------------------------
# Save all data currently in memory.
# First writes the name of the button and then
# a sequence of hex numbers + a line break at the
# end. Every line is a button
#
# MISSING: should save the carrier frequence the IR 
#          data was recorded with (and possibly other 
#          config information)
#---------------------------------------------------------------------
proc Slinke::save {{filename slinke.buttons}} {
  set fout [open $filename w]
  foreach f $Slinke::l_buttonnames {
    #puts "Saving button $f"
    set len [string length $Slinke::a_buttons($f)]
    set b $Slinke::a_buttons($f)
    set msg ""
    for {set i 0} {$i < $len} {incr i} {
      scan [string range $b $i $i] "%c" c
      append msg [format "/%02x" $c]
    }
    
    puts $fout [list $f $msg]
  }
  close $fout
}


#-----------------------------------------------------------------------
# Slinke::load --- load previously stored buttons and IR messages
#
# Load a previously stored file with buttons and
# their IR data and put all that information into
# the memory for Slinke::recall's to work.
#-----------------------------------------------------------------------
proc Slinke::load {{filename slinke.buttons}} {
  if {[catch {set fin [open $filename r]}]} then {
    puts "File $filename could not be found."
    return "File not found"
  }
  if {"" != $fin} then {
    while {1} {
      # Read a line
      set error 0
      gets $fin line
      if {"" != $line} {
        set button [lindex $line 0]
        #puts "read button $button"
        set msg [lindex $line 1]
        #puts "msg=$msg"
        set bmsg ""
        set len [string length $msg]
        for {set i 0} {$i < $len} {set i [expr $i+3]} {
          set sub [string range $msg [expr $i+1] [expr $i+2]]
          set r [scan $sub "%x" c]
          if {1 != $r} then {
            set error 1
            break
          }
          append bmsg [format "%c" $c]
        }
        if {0 == $error} then {
          Slinke::store $button $bmsg ""
        }
      } else {
        break
      }
    }
  }
}


#------------------------------------------------------------------------
# Slinke::learnsession --- learn commands and store them
#
# Run a learn session. The user must enter the
# names of the buttons with the keyboard and
# after it is finished all the data is stored
# in the default file.
#------------------------------------------------------------------------
proc Slinke::learnsession {{filename slinke.buttons}} {
  while {1} {
    puts "Enter name of button to learn or '--' to exit"
    gets stdin input
    if {"--" == $input} {
      break
    }
    set rc [Slinke::learn "IR"]
    puts "Storing message under button with name $input ."
    Slinke::store $input $rc
  }

  puts "Saving buttons in file $filename."
  Slinke::save $filename
}


#-------------------------------------------------------------------------
# Slinke::ir --- test function 'instant replay'
#
# Immediately sends a command received over IR to the IR transmitter
#-------------------------------------------------------------------------
proc Slinke::ir {} {
  while {1} {
    set rc [Slinke::learn IR]
    Slinke::play IR $rc
  }
}

proc Slinke::test {} {
  puts "Version: [Slinke::get_version]"
  puts "Serial number: [Slinke::get_serial]"

  puts "Baudrate: [Slinke::get_baudrate]"

  puts "Sample period [Slinke::get_ir_sample_period] microseconds"
  puts "Setting sample period to 100 microseconds...[Slinke::set_ir_sample_period 100]"
  puts "Sample period [Slinke::get_ir_sample_period] microseconds"

  puts "IR carrier: [Slinke::get_ir_carrier]"

  puts "IR timeout [Slinke::get_ir_timeout]"
  puts "Setting IR timeout to 600... [Slinke::set_ir_timeout 600]"
  puts "IR timeout [Slinke::get_ir_timeout]"

  puts "Minimum IR message length [Slinke::get_min_ir_msg_len]"
  puts "Setting minimum IR message length to 15... [Slinke::set_min_ir_msg_len 15]"
  puts "Minimum IR message length [Slinke::get_min_ir_msg_len]"

  puts "IR Transmit ports [Slinke::get_ir_transmit_ports]"
  puts "Setting IR Transmit ports to 1 ... [Slinke::set_ir_transmit_ports 1]"
  puts "IR Transmit ports [Slinke::get_ir_transmit_ports]"

  puts "IR Receive port [Slinke::get_ir_receive_ports]"
  puts "Setting IR Receive port to 255... [Slinke::set_ir_receive_ports 255]"
  puts "IR Receive port [Slinke::get_ir_receive_ports]"

  puts "IR Receive port polarities [Slinke::get_ir_receive_port_polarities]"
  puts "Setting IR Receive port polarities to 15... [Slinke::set_ir_receive_port_polarities 15]"
  puts "IR Receive port polarities [Slinke::get_ir_receive_port_polarities]"

  puts "IR Routing table [Slinke::get_ir_routing]"

  puts "Parallel port handshaking mode [Slinke::get_handshaking]"
  puts "Setting handshaking mode to 7... [Slinke::set_handshaking 7]"
  puts "Parallel port handshaking mode [Slinke::get_handshaking]"

  puts "Parallel port direction [Slinke::get_direction]"
  puts "Setting direction to 0... [Slinke::set_direction 0]"
  puts "Parallel port direction [Slinke::get_direction]"
}

Local::init
Slinke::init
#Slinke::connect "/dev/ttyS0"
#Slinke::set_baudrate 19200
#Slinke::disconnect

package provide slinke 0.8
